-- Util functions, called by dddl code
--
-- Note that the utilBuff, utilDebuff, and utilDebuffByMe routines all 
-- make the assumption that the spell that casts the buff/debuff and the 
-- buff/debuff have the same name, so it checks cast time by the buff name.  
-- This is a hard problem to fix because some buffs should consider cast time
-- (normal buffs) and some shouldn't (proc'ed buffs, like Nightfall).  As
-- a practical matter, every castable buff/debuff seems to have the same 
-- name as the spell, so this rather embarassing bug is not a bad kludge
-- to deal with all cases (proc'd buffs will return a 0 cast time so won't
-- be clipped).  At some point I should fix it properly.

-- return true if the passed buff is up on the doofus, and will extend past the
-- clip time (a clip of 0 is no clip) plus the cast time.  False otherwise
function DestroDummies:utilBuff(buff, clip, doofus)
    local time = DestroDummies:buffTime(buff);

    local name, _,_, _, _, _, expiry, _, _ = UnitAura(doofus, buff, HELPFUL);

    if (name == nil) then
        return false;
    end

    if (expiry == nil or expiry == 0) then
        -- no expiry means that the buff just stays up (cat form, aspects, etc)
        -- note that a 0 expiry means time 0, which is the computer's boot time
        -- so it's not really ever a valid time
        return true;
    end

    -- Get the cast time.  This *cannot* be cached because it will change as
    -- the player's haste changes, which can be triggered by items
    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(buff);  -- in milliseconds
    if (castTime == nil) then
    	-- Checks for proc'cing buffs will have a nil cast time
    	castTime = 0;
    end
    castTime = castTime / 1000;		-- now in seconds

    return (expiry - GetTime() - castTime) > clip;
end


-- return the amount of time left on the buff, or 0 if it does not exist.
-- NB: permanent buffs (cat form, hunter aspects, etc) will return 0 when
-- they are up also.  Using this routine presupposes that you know that the
-- spell has a time associated with it.
--
-- Note also that this returns true, not clipped, time, even if you have a 
-- clip specified for this spell
function DestroDummies:utilBuffTime(buff)
    local name, _,_, _, _, _, expiry, _, _ = UnitAura("player", buff, HELPFUL);

    if (name == nil or expiry == 0) then
        return 0;
    else
        return expiry - GetTime();
    end
end


-- return true if the passed debuff is up by anyone (and will extend past the
-- clip plus cast time), false otherwise.  A clip of 0 means no clip
function DestroDummies:utilDebuff(debuff, clip)
    if (clip == nil) then
    	clip = 0;
    end

    local name, expiry = DestroDummies:debuffUpByAny(debuff);
    if (name == nil) then
        return false;
    end

    -- Get the cast time.  This *cannot* be cached because it will change as
    -- the player's haste changes, which can be triggered by items
    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(debuff);  -- in milliseconds
    castTime = castTime / 1000;		-- now in seconds

    return (expiry - GetTime() - castTime) > clip;
end

-- return true if the passed debuff is up by me (and will extend past the
-- clip plus cast time), false otherwise.  A clip of 0 means no clip
function DestroDummies:utilDebuffByMe(debuff, clip)
    local name, expiry = DestroDummies:debuffUpByMe(debuff);
    
    if (name == nil) then
        return false;
    end

    -- Get the cast time.  This *cannot* be cached because it will change as
    -- the player's haste changes, which can be triggered by items
    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(debuff);  -- in milliseconds
    castTime = castTime / 1000;		-- now in seconds

    return (expiry - GetTime() - castTime) > clip;
end

-- return the seconds left on the buff/debuff (or 0 if it's not up), and the
-- count of the buff/debuff stack (or 0 if it's not up)
function DestroDummies:utilBuffInfo(buff, isDebuff)
    local name, expiry, count;
    if (isDebuff) then
	name, expiry, count = DestroDummies:debuffUpByAny(buff);
	if (expiry ~= nil) then
	    expiry = expiry - GetTime();
	end
    else
        expiry, count = DestroDummies:buffTime(buff);
	name = buff;
    end

    if (name == nil or expiry == nil or expiry == 0) then
        return 0, 0;
    else
        return expiry, count;
    end
end
    
-- check a mutex to see if any of the buffs are up on the player.  if a buff
-- is castable by the player, consider it not up if it is within the clip
-- time.  Return the name of the buff that is up, or nil if none
function DestroDummies:utilMxBuff(mNum)
    for i = 1,DestroDummies.mutex[mNum].count do
    	local name, _,_, _, _, _, expiry, caster, _ = 
		    UnitAura("player", DestroDummies.mutex[mNum][i], HELPFUL);

        if (name ~= nil) then
	    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(name);
	    -- We get nil if the buff is not the one we can cast (for example,
	    -- Abominable Might having overwritten Trueshot Aura on a Hunter)
	    if (castTime == nil) then
		return nil;
	    end
	    castTime = castTime / 1000;		-- milliseconds to seconds

	    -- pick the right clip to use
	    local clip;
	    if (UnitAffectingCombat("player")) then
	    	clip = DestroDummies.mutex[mNum].clipCombat;
	    else
	    	clip = DestroDummies.mutex[mNum].clipOOCombat;
	    end

	    if (expiry == 0) or 
	       ((expiry - GetTime() - clip - castTime) > 0) then
		return name;
	    else
	    	return nil;
	    end
        end
    end
    return nil;
end

-- check a mutex for which of the buffs is up on the player (cast by him, 
-- to catch the case where another hunter has Aspect of the Pack up)
function DestroDummies:utilMxBuffByMe(mNum)
    for i = 1,DestroDummies.mutex[mNum].count do
    	local name, _,_, _, _, _, expiry, caster, _ = 
		    UnitAura("player", DestroDummies.mutex[mNum][i], HELPFUL);

        if (name ~= nil and caster == "player") then
	    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(name);
	    -- This should never happen
	    if (castTime == nil) then
		return nil;
	    end
	    castTime = castTime / 1000;		-- milliseconds to seconds

	    -- pick the right clip to use
	    local clip;
	    if (UnitAffectingCombat("player")) then
	    	clip = DestroDummies.mutex[mNum].clipCombat;
	    else
	    	clip = DestroDummies.mutex[mNum].clipOOCombat;
	    end

	    if (expiry == 0) or 
	       ((expiry - GetTime() - clip - castTime) > 0) then
		return name;
	    else
	    	return nil;
	    end
        end
    end
    return nil;
end

-- check a mutex for which of the debuffs is up on the target (cast by 
-- the player)
function DestroDummies:utilMxDebuffByMe(mNum)
    for i = 1,DestroDummies.mutex[mNum].count do
        local name, expiry = DestroDummies:debuffUpByMe(DestroDummies.mutex[mNum][i]);
        if (name ~= nil) then
	    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(name);
	    castTime = castTime / 1000;		-- milliseconds to seconds
	    expiry = expiry - GetTime();

	    -- pick the right clip to use
	    local clip;
	    if (UnitAffectingCombat("player")) then
	    	clip = DestroDummies.mutex[mNum].clipCombat;
	    else
	    	clip = DestroDummies.mutex[mNum].clipOOCombat;
	    end

	    if (expiry > clip + castTime) then
	    	return name;
	    else
	    	return nil;
	    end
        end
    end
    return nil;
end

-- Check a mutex to see if we should apply the passed-in debuff to the 
-- target.  We should if there is no higher-level debuff (ie., declared 
-- earlier) already on, or if expiry is less than the clip time.  Return
-- true if we should cast, false if not
function DestroDummies:utilMxNeedDebuff(mNum, spell)
    -- loop on each spell in the mutex, in decreasing priority
    for i = 1,DestroDummies.mutex[mNum].count do
        local name, expiry, count = DestroDummies:debuffUpByAny(DestroDummies.mutex[mNum][i]);
        if (name == nil) then
	    -- the debuff is not up.  If it is us, we should apply it
	    if (spell == DestroDummies.mutex[mNum][i]) then
	    	return true;
	    end
	else
	    -- We have a debuff.  If it's not the spell we are looking
	    -- for, there is a higher priority debuff up, so we we should
	    -- not cast
	    if (name == spell) then
	    	return false;
	    end

	    -- It's us.  Check the expiry
	    local _,_,_,_,_,_,castTime,_,_ = GetSpellInfo(name);
	    castTime = castTime / 1000;		-- milliseconds to seconds
	    expiry = expiry - GetTime();

	    -- pick the right clip to use
	    local clip;
	    if (UnitAffectingCombat("player")) then
	    	clip = DestroDummies.mutex[mNum].clipCombat;
	    else
	    	clip = DestroDummies.mutex[mNum].clipOOCombat;
	    end

	    return (expiry > clip + castTime);
        end
    end
    return true;
end

-- return true if the passed weapon has an enhancement, false otherwise
function DestroDummies:utilWeaponEnhance(clip)
    local mainChant, mainTime, _, offChant, offTime, _ = GetWeaponEnchantInfo();

    clip = clip * 1000;		-- convert to milliseconds
    if (mainChant == nil or mainTime < clip) then
    	return false;
    end
    if (OffhandHasWeapon == 1 and (offChant == nil or offTime < clip)) then
	return false;
    end
    return true;
end

-- return the passed things health, as a percent of max.  if the passed thing
-- does not exist, return that it's at full health (so tests on pets that 
-- aren't out don't result in constant calls to heal the pet
function DestroDummies:utilHealth(thang)
    if not UnitExists(thang) then
    	return 100;
    end

    local cur = UnitHealth(thang);
    local max = UnitHealthMax(thang);

    if (cur == nil or max == nil or max == 0) then
    	return 0;
    end
    return (cur * 100) / max;
end

-- Extrapolate how long it will take to kill the target
function DestroDummies:utilTimeToDie()
    local perc = DestroDummies:utilHealth("target");
    local elapsed = GetTime() - DestroDummies.combatStartTime;
    
    return (elapsed * 100) / (100 - perc);
end

-- UnitDetailedThreatSituation() tends to return nil a lot, especially
-- when practicing on the dummies in Ironforge.  So turn them all into 0's
-- for our nice users
function DestroDummies:utilThreat()
    local isTanking, status, threatpct, rawthreatpct, threatvalue = 
		UnitDetailedThreatSituation("player", "target");
    if isTanking == nil then
	isTanking = 0;
    end
    if status == nil then
	status = 0;
    end
    if threatpct == nil then
	threatpct = 0;
    end
    if rawthreatpct == nil then
	rawthreatpct = 0;
    end
    if threatvalue == nil then
	threatvalue = 0;
    end
    return isTanking, status, threatpct, rawthreatpct, threatvalue;
end

function DestroDummies:utilHasGlyph(name)
    for i = 1, 6 do
        local _, _, glyphSpell, _ = GetGlyphSocketInfo(i);
	if (glyphSpell ~= nil) then
	    local gname,_,_,_,_,_,_,_,_ = GetSpellInfo(glyphSpell);
	    if (gname == name) then
	    	return true;
	    end
	end
    end
    return false;
end

-- Return the amount of time it will take this spell to fly the distance,
-- *assuming* it actually does have to fly and is not instant.  Knowledge
-- of which spells need to call this rests with you, dear programmer, not
-- with me
--
-- Assume a flight time of 20yds/sec, which is what I timed on a stopwatch
--
function DestroDummies:utilFlightTime()
    return DestroDummies:distanceToTarget() / 20.0;
end


-- Something is a boss if it is elite and 3 or more levels above you
function DestroDummies:utilBoss()
    local uc = UnitClassification("target");

    return ((uc == "worldboss" or uc == "rareelite" or uc == "elite") and
            (UnitLevel("target") >= (UnitLevel("player") + 3)));
end


-- For now, just return true if the passed totem type is deployed.  This is 
-- just stub functionality.
function DestroDummies:utilTotem (totType)
    local haveTotem, totemName, startTime, duration = GetTotemInfo(totType);
    return startTime ~= 0;
end
